/*
 * gscodec.c
 *
 * Simple GSLib COmpression/DECompression program.
 * Don`t forget to copy gs.dll into the working
 * directory before runing!
 *
 * (C) ENTROPYWARE, Alexander Simakov.
 *
 * Our site: http://www.entropyware.info/
 * Support:  support@entropyware.info
 * Author:   xander@entropyware.info
 */

#include <stdio.h>
#include <stdlib.h>
#include "pbm.h"   /* PGM/PPM support */
#include "gslib.h" /* GSLib interface */

#define BUFFER_SIZE (4096)

/* User supplied memory allocation function */
void *GS_API my_alloc(int size, int items)
{
  return malloc(size * items);
}

/* User supplied memory releasing function */
void GS_API my_free(void *address)
{
  free(address);
}

/* User supplied progress reporting function */
void GS_API my_progress(int percent_done)
{
  fprintf(stderr, "\rProgress: %d%% done", percent_done);
  fflush(stderr);
}

void compress_file(char *in_path, char *out_path, int quality)
{
  FILE *in_file, *out_file;
  char *in_buf, *out_buf;
  gs_stream gs_strm;
  int response, bytes_read;
  pbm_hdr hdr;

  /* Open files */
  if (((in_file = fopen(in_path, "rb")) == NULL) || ((out_file = fopen(out_path, "wb")) == NULL)) {
    printf("Can`t open/create file(s).\n");
    exit(1);
  }

  /* Allocate buffers */
  if (((in_buf = (char *) malloc(BUFFER_SIZE)) == NULL) || ((out_buf = (char *) malloc(BUFFER_SIZE)) == NULL)) {
    printf("Not enough memory.\n");
    exit(1);
  }

  /* Read PGM or PPM file header */
  if (read_hdr(in_file, &hdr)) {
    printf("Corrupted file or usupported format.\n");
    exit(1);
  }

  /* Prepare GSLib */
  gs_strm.next_in = in_buf;
  gs_strm.avail_in = 0;

  gs_strm.next_out = out_buf;
  gs_strm.avail_out = 0;

  gs_strm.img_width = hdr.width;
  gs_strm.img_height = hdr.height;

  gs_strm.quality = quality;

  gs_strm.gs_alloc = my_alloc;
  gs_strm.gs_free = my_free;
  gs_strm.gs_progress = my_progress;

  /* Start GSLib */
  if (hdr.type == PGM) response = gs_start_compress(&gs_strm, GS_GRAY); /* grayscale PGM */
  else response = gs_start_compress(&gs_strm, GS_RGB);                  /* truecolor PPM */

  if (response != GS_OK) {
    printf("GSLib initialization fail.\n");
    exit(1);
  }

  /* Main compression loop */

  do {

    /* Compress next portion of data */    
    response = gs_compress(&gs_strm);

    switch (response) {

      /* Input buffer is empty, read more input */
      case GS_BUFFER_EMPTY:
      bytes_read = fread(in_buf, 1, BUFFER_SIZE, in_file);
      gs_strm.next_in = in_buf;
      gs_strm.avail_in = bytes_read;
      break;

      /* Output buffer is full, write more output */
      case GS_BUFFER_FULL:
      fwrite(out_buf, 1, gs_strm.next_out - out_buf, out_file);
      gs_strm.next_out = out_buf;
      gs_strm.avail_out = BUFFER_SIZE;
      break;

      /* Work complete, flush remaining output */
      case GS_OK:
      fwrite(out_buf, 1, gs_strm.next_out - out_buf, out_file);
      break;
    }
  } while (response != GS_OK); /* That`s it! */

  /* Stop GSLib */
  gs_end_compress(&gs_strm);

  fclose(in_file); fclose(out_file);
  free(in_buf); free(out_buf);
}

void decompress_file(char *in_path, char *out_path)
{
  FILE *in_file, *out_file;
  char *in_buf, *out_buf;
  gs_stream gs_strm;  
  int response, bytes_read;
  
  /* Open files */
  if (((in_file = fopen(in_path, "rb")) == NULL) || ((out_file = fopen(out_path, "wb")) == NULL)) {
    printf("Can`t open/create file(s).\n");
    exit(1);
  }

  /* Allocate buffers */
  if (((in_buf = (char *) malloc(BUFFER_SIZE)) == NULL) || ((out_buf = (char *) malloc(BUFFER_SIZE)) == NULL)) {
    printf("Not enough memory.\n");
    exit(1);
  }

  /* Prepare GSLib */
  gs_strm.next_in = in_buf;
  gs_strm.avail_in = 0;

  gs_strm.next_out = out_buf;
  gs_strm.avail_out = 0;

  gs_strm.gs_alloc = my_alloc;
  gs_strm.gs_free = my_free;
  gs_strm.gs_progress = my_progress;

  /* Start GSLib */
  response = gs_start_decompress(&gs_strm, file_size(in_file));

  if (response != GS_OK) {
    printf("GSLib initialization fail.\n");
    exit(1);
  }

  /* Main decompression loop */

  do {

    /* Decompress next portion of data */
    response = gs_decompress(&gs_strm);

    switch (response) {

      /* Input buffer is empty, read more input */
      case GS_BUFFER_EMPTY:
      bytes_read = fread(in_buf, 1, BUFFER_SIZE, in_file);
      gs_strm.next_in = in_buf;
      gs_strm.avail_in = bytes_read;
      break;

      /* Output buffer is full, write more output */
      case GS_BUFFER_FULL:
      fwrite(out_buf, 1, gs_strm.next_out - out_buf, out_file);
      gs_strm.next_out = out_buf;
      gs_strm.avail_out = BUFFER_SIZE;
      break;

      /* GS header is damaged */
      case GS_HEADER_FAIL:
      printf("The GS header is damaged.\n");
      gs_end_decompress(&gs_strm);
      exit(1);
      break;

      /* Not enough memory */
      case GS_MEMORY_FAIL:
      printf("Not enough memory.\n");
      gs_end_decompress(&gs_strm);
      exit(1);
      break;

      /* GS header is successfully read, write PGM header */
      case GS_HEADER_OK:
      fprintf(out_file, "P5\n%d %d\n255\n", gs_strm.img_width, gs_strm.img_height);
      break;

      /* Work complete, flush remaining output */
      case GS_OK:
      fwrite(out_buf, 1, gs_strm.next_out - out_buf, out_file);
      break;
    }
  } while (response != GS_OK); /* That`s it! */

  /* Stop GSLib */
  gs_end_decompress(&gs_strm);

  fclose(in_file); fclose(out_file);
  free(in_buf); free(out_buf);
}

int main(int argc, char *argv[])
{
  printf("\nGSCODEC, (C) ENTROPYWARE, Alexander Simakov.\n");

  if ((argc == 5) && (*argv[1] == 'c') && (atoi(argv[2]) >= 0) && (atoi(argv[2]) <= 100)) {
    compress_file(argv[3], argv[4], atoi(argv[2]));
    printf("\n");
    return 0;
  }

  if ((argc == 4) && (*argv[1] == 'd')) {
    decompress_file(argv[2], argv[3]);
    printf("\n");
    return 0;
  }

  printf("\nUsage: <gscodec.exe> <c|d> [0<=quality<=100] <in_file> <out_file>\n\n");
  printf("Example 1: gscodec.exe c 65 lena.pgm lena.gs\n");
  printf("Example 2: gscodec.exe c 20 cat.ppm cat.gs\n");
  printf("Example 3: gscodec.exe d goldhill.gs goldhill.pgm\n");

  return 1;
}
