/*
 * gslib.h
 *
 * The GrayScale, DCT-based lossy image compression library.
 *
 * (C) ENTROPYWARE, Alexander Simakov.
 *
 * Our site: http://www.entropyware.info/
 * Support:  support@entropyware.info
 * Author:   xander@entropyware.info
 */

#ifndef GSLIB_H
#define GSLIB_H

#define GS_API __stdcall

#ifdef __cplusplus
extern "C" {
#endif

/* The GrayScale compression library provides in-memory compression and
 * decompression functions.
 *
 * Compression (decompression) can be done in a single step if the buffers
 * are large enough, or can be done by repeated calls of the compression
 * (decompression) function.  In the latter case, the application must
 * provide more input and/or consume the output (providing more output
 * space) before each call.
 *
 * The library does not install any signal handler. Although the decoder
 * checks only the consistency of the header, the library should never
 * crash: due to lossy compression, errors in the data stream does not
 * matter (however a damaged output image is likely to result).
 */

/* Return codes for the compression/decompression functions.
 * Negative values are errors, positive values are used for
 * special but normal events.
 */

#define GS_BUFFER_FULL  (3)  /* output buffer is full */
#define GS_BUFFER_EMPTY (2)  /* input buffer is empty */
#define GS_HEADER_OK    (1)  /* the header is successfully read */
#define GS_OK           (0)  /* work is completed */
#define GS_PARAMS_FAIL  (-1) /* parameters are incorrect */
#define GS_MEMORY_FAIL  (-2) /* there is not enough memory */
#define GS_HEADER_FAIL  (-3) /* the header is damaged */

/* Input color space & format */

#define GS_GRAY "GRAY" /* 8-bit grayscale format */
#define GS_RGB  "RGB"  /* 24-bit truecolor format */

/* If your application wants to deal with data in the order B,G,R, just
 * use "BGR". You can also deal with formats such as R,G,B,X (one extra
 * byte per pixel) by using "RGBX". Some examples below.
 */

#define GS_BGR  "BGR"
#define GS_RGBX "RGBX"
#define GS_XBGR "XBGR"

/* GS stream structure */

typedef struct
{
  char *next_in;   /* next input byte */
  int  avail_in;   /* number of bytes available at next_in */
  int  total_in;   /* total number of input bytes read so far */

  char *next_out;  /* next output byte should be put there */
  int  avail_out;  /* remaining free space at next_out */
  int  total_out;  /* total number of bytes output so far */

  int  img_width;  /* image width */
  int  img_height; /* image height */

  int  quality;    /* quality level */

  void *state;     /* internal compression/decompression state */

  void *(GS_API *gs_alloc)(int size, int items); /* memory allocation function */
  void (GS_API *gs_free)(void *address);         /* memory releasing function */
  void (GS_API *gs_progress)(int percent_done);  /* progress report function */
} gs_stream;

/* The application must update next_in and avail_in when avail_in has
 * dropped to zero. It must update next_out and avail_out when avail_out
 * has dropped to zero.
 *
 * The fields total_in and total_out can be used for statistics. Those
 * fields are set (and used) by the compression library and must not be
 * updated by the application.
 *
 * The field state is not visible by applications.
 *
 * If the GS library is used in a multi-threaded application, gs_alloc,
 * gs_free and gs_progress must be thread safe.
 *
 * To avoid progress reporting set gs_progress to NULL.
 */

int GS_API gs_start_compress(gs_stream *s, char *fmt);

/* Initializes the internal stream state for compression. The fields
 * gs_alloc, gs_free, gs_progress, img_width, img_height and quality
 * must be initialized before by the caller.
 *
 * The fields img_width and img_height must be between 1 and 16384.
 *
 * The field quality must be between 0 and 100. Increasing quality
 * degradates compression ratio, decreasing quality favors size.
 * The quality of 65 works fine in most cases.
 *
 * fmt is a null-terminated string, that contain input color space
 * and format (see above).
 *
 * gs_start_compress returns GS_OK if success, GS_MEMORY_FAIL if there
 * was not enough memory or GS_PARAMS_FAIL if the input parameters were
 * incorrect.
 *
 * gs_start_compress does not perform any compression: this will be
 * done by gs_compress.
 */

int GS_API gs_compress(gs_stream *s);

/* gs_compress compresses as much data as possible, and stops when the input
 * buffer becomes empty, the output buffer becomes full or the image was
 * compressed.
 *
 * The fields next_in, avail_in, next_out and avail_out must be set before
 * by the caller.
 *
 * gs_compress returns GS_BUFFER_EMPTY if the input buffer becomes empty
 * (i.e. avail_in == 0). In this case the application must provide more input
 * and update next_in and avail_in accordingly.
 *
 * gs_compress returns GS_BUFFER_FULL if the output buffer becomes full
 * (i.e. avail_out == 0). In this case the application must consume the
 * output (providing more output space) and update next_out and avail_out
 * accordingly.
 *
 * The application can consume the compressed output when it wants, for
 * example when the output buffer is full (avail_out == 0), or after each
 * call of gs_compress.
 *
 * gs_compress returns GS_OK if the image was compressed. The application
 * must flush pending output.
 */

void GS_API gs_end_compress(gs_stream *s);

/* All dynamically allocated data structures for this stream are freed.
 * This function discards any unprocessed input and does not flush any
 * pending output.
 */

int GS_API gs_start_decompress(gs_stream *s, int stream_size);

/* Initializes the internal stream state for decompression. The fields
 * gs_alloc, gs_free, gs_progress, must be initialized before by the caller.
 *
 * stream_state must be set to compressed stream size.
 *
 * gs_start_decompress returns GS_OK if success, GS_MEMORY_FAIL if there
 * was not enough memory or GS_PARAMS_FAIL if the input parameters were
 * incorrect.
 *
 * gs_start_decompress does not perform any decompression: this will be
 * done by gs_decompress.
 */

int GS_API gs_decompress(gs_stream *s);

/* gs_decompress decompresses as much data as possible, and stops when
 * the input buffer becomes empty, the output buffer becomes full or
 * the image was decompressed.
 *
 * The fields next_in, avail_in, next_out and avail_out must be set before
 * by the caller.
 *
 * gs_decompress returns GS_BUFFER_EMPTY if the input buffer becomes empty
 * (i.e. avail_in == 0). In this case the application must provide more input
 * and update next_in and avail_in accordingly.
 *
 * gs_decompress returns GS_BUFFER_FULL if the output buffer becomes full
 * (i.e. avail_out == 0). In this case the application must consume the
 * output (providing more output space) and update next_out and avail_out
 * accordingly.
 *
 * The application can consume the uncompressed output when it wants, for
 * example when the output buffer is full (avail_out == 0), or after each
 * call of gs_decompress.
 *
 * gs_decompress returns GS_HEADER_FAIL if the header was damaged.
 * Future decompression is impossible.
 *
 * gs_decompress returns GS_HEADER_OK if the header was successfully read.
 * The library will set the fields img_width, img_height and quality
 * accordingly.
 *
 * gs_decompress returns GS_MEMORY_FAIL if there was not enough memory.
 * Future decompression is impossible.
 *
 * gs_decompress returns GS_OK if the image was decompressed. The application
 * must flush pending output.
 */

void GS_API gs_end_decompress(gs_stream *s);

/* All dynamically allocated data structures for this stream are freed.
 * This function discards any unprocessed input and does not flush any
 * pending output.
 */

#ifdef __cplusplus
}
#endif

#endif /* GSLIB_H */
