/*
 * pbm2bmp.c
 *
 * PGM & PPM -> BMP conversion utility.
 *
 * (C) ENTROPYWARE, Alexander Simakov.
 *
 * Our site: http://www.entropyware.info/
 * Support:  support@entropyware.info
 * Author:   xander@entropyware.info
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "pbm.h"

/* Mirror a pixel: BGR -> RGB */
#define MIRROR(bgr) (temp = bgr[0], bgr[0] = bgr[2], bgr[2] = temp)

void pbm2bmp(char *in_path, char *out_path)
{
  FILE *in_file, *out_file;
  BITMAPFILEHEADER bmp_hdr1;
  BITMAPINFOHEADER bmp_hdr2;
  pbm_hdr hdr;
  int i, bmp_size, scanline_size, pad_bytes, pixel_size, rgb_val, cur_pos;
  int width, height;
  char *scanline, *next_pixel, temp;

  /* Open files */
  if (((in_file = fopen(in_path, "rb")) == NULL) || ((out_file = fopen(out_path, "wb")) == NULL)) {
    printf("Can`t open/create file(s).\n");
    exit(1);
  }

  /* Read PGM or PPM file header */
  if (read_hdr(in_file, &hdr)) {
    printf("Corrupted file or usupported format.\n");
    exit(1);
  }

  width = hdr.width;
  height = hdr.height;
  pixel_size = (hdr.type == PGM ? 1 : 3);

  /* Compute scanline size, including padding to 4-byte boundary */
  scanline_size = width * pixel_size;
  if (scanline_size & 3) scanline_size = (scanline_size & 0xfffffffc) + 4;

  /* Compute BMP size */ 
  bmp_size = 14 + 40 + (hdr.type == PGM ? 1024 : 0) + scanline_size * height;

  pad_bytes = scanline_size - width * pixel_size;

  if ((scanline = (char *) malloc(scanline_size)) == NULL) {
    printf("Not enough memory.\n");
    exit(1);
  }

  /* Fill BITMAPFILEHEADER */
  bmp_hdr1.bfType = 0x4d42;
  bmp_hdr1.bfSize = bmp_size;
  bmp_hdr1.bfReserved1 = 0;
  bmp_hdr1.bfReserved2 = 0;
  bmp_hdr1.bfOffBits = 14 + 40 + (hdr.type == PGM ? 1024 : 0);

  /* Fill BITMAPINFOHEADER */
  bmp_hdr2.biSize = 40;
  bmp_hdr2.biWidth = width;
  bmp_hdr2.biHeight = height;
  bmp_hdr2.biPlanes = 1;
  bmp_hdr2.biBitCount = (WORD) (hdr.type == PGM ? 8 : 24);
  bmp_hdr2.biCompression = 0;
  bmp_hdr2.biSizeImage = 0;
  bmp_hdr2.biXPelsPerMeter = 0;
  bmp_hdr2.biYPelsPerMeter = 0;
  bmp_hdr2.biClrUsed = 0;
  bmp_hdr2.biClrImportant = 0;

  /* Write BMP headers */
  fwrite(&bmp_hdr1, 1, 14, out_file);
  fwrite(&bmp_hdr2, 1, 40, out_file);

  /* Set linear 8-bit grayscale palette(for PGM files) */
  if (hdr.type == PGM) {
    for (i = 0; i < 256; i++) {
      rgb_val = ((i << 16) | (i << 8) | (i));
      fwrite(&rgb_val, 1, 4, out_file);
    }
  }

  /* Set file pointer to the last scanline */
  fseek(out_file, bmp_size - scanline_size, SEEK_SET);

  /* Since BMP stores scanlines bottom-to-top, we have to invert the image
   * from PGM/PPM top-to-bottom order. To do this, we write the outcoming data
   * in reverse order.
   */

  while (height > 0) {
    fread(scanline, 1, scanline_size - pad_bytes, in_file);

    /* BMP stores pixels in reverse order: BGR instead of RGB */
    if (pixel_size == 3) {
      for (i = 0, next_pixel = scanline; i < width; i++) {
        MIRROR(next_pixel);
        next_pixel += 3;
      }
    }

    fwrite(scanline, 1, scanline_size, out_file);
    cur_pos = ftell(out_file);
    fseek(out_file, cur_pos - 2 * scanline_size, SEEK_SET);
    height--;
  }

  fclose(in_file); fclose(out_file);
  free(scanline);
}

int main(int argc, char *argv[])
{
  printf("\nPGM & PPM -> BMP conversion utility, (C) ENTROPYWARE, Alexander Simakov.\n");

  if (argc == 3) {
    pbm2bmp(argv[1], argv[2]);
    printf("\n");
    return 0;
  }

  printf("\nUsage: <pbm2bmp.exe> <in_file> <out_file>\n\n");
  printf("Example 1: pbm2bmp.exe lena.pgm lena.bmp\n");
  printf("Example 2: pbm2bmp.exe cat.ppm cat.bmp\n");

  return 1;
}
